#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <sys/wait.h>

/* Amount of time an entry will stay in database in seconds */
#define GRACE (60 * 30)

/* Number of entries the database will hold */
#define NUM_ENTRIES	(250)

/* Change these to point to your log file and where the */
/* pop-before-smtp "database" should be stored          */
#define POP_BEFORE_SMTP_DIR "/etc/postfix"
#define POP_LOG "/var/log/mail/info"

/* Change to define which pop3 deamon you are using */
/* VMPOP3D  QPOPPER IPOP3D TEAPOP VDPOP3D */
/* COURIER TPOP3D CYRUS */
#define VMPOP3D

/* Uncomment below line if you want to turn Optimization settings on */
/*#define OPTIMIZE*/
#define SCAN_DEPTH	10

#define VERSION "1.0b"

/* head always points to the next available entry */
int head=0, tail=0, debug = 0, rewrite = 1;
pid_t child_id;
char pop_before_smtp_dir[80] = POP_BEFORE_SMTP_DIR;

struct database {
	char ip_addr[20];
	time_t expire;
} pop_ok[NUM_ENTRIES];

#ifdef VMPOP3D
inline char* match(char *ptr)
{
	char *str_ptr;
	
	if (strstr(ptr, "vm-pop3d") != NULL ) {	
		str_ptr = strstr(ptr, "logged in from");
			
		if (str_ptr != NULL) {
			str_ptr += strlen("logged in from ");
			
			return str_ptr;
		}
	}

	return NULL;
}
#endif

#ifdef QPOPPER
inline char* match(char *ptr)
/* Submitted by Anton Krall 4/1/02 */
/* akrall@team.inter.net or netlord@axtel.net */
/* Changed slightly for different qpopper log */
/* Log submitted by Felipe Ceglia 11/23/02 */
/* felipe@terenet.com.br 		*/
{
	char *str_ptr;
	
	if (strstr(ptr, "popper") != NULL ) {	
		str_ptr = strstr(ptr, "))");
			
		if (str_ptr != NULL) {
			str_ptr += strlen(")) ");
			
			return str_ptr;
		}
	}

	return NULL;
}
#endif

#ifdef IPOP3D
/* Code to check for Auth Submitted by     */
/* Kevin Brokaw 8/7/02 kevinb@moscorp.net  */
inline char* match(char *ptr)
{
	char *str_ptr;
	
	if (strstr(ptr, "ipop3d") != NULL ) {	
		str_ptr = strstr(ptr, "Login user");
			
		if (str_ptr != NULL) {
			str_ptr = strstr(str_ptr, "[");
			str_ptr++;
			
			return str_ptr;
		}
		
		str_ptr = strstr(ptr, "Auth user");

		if (str_ptr != NULL) {
			str_ptr = strstr(str_ptr, "[");
			str_ptr++;
			
			return str_ptr;
		}
	}

	return NULL;
}
#endif

#ifdef TEAPOP
inline char* match(char *ptr)
/* Submitted by Skip Watson 6/18/02 */
/* ciaran@aldhfn.org */
{
	char *str_ptr;
   if (strstr(ptr, "teapop") != NULL ) {	
   	str_ptr = strstr(ptr, "from");
   	
   	if (str_ptr != NULL) {
   		str_ptr = strstr(str_ptr, "[");
   	 	str_ptr++;
   	 	
   	 	return str_ptr;
		}
   }
   
   return NULL;
}
#endif

#ifdef VDPOP3D
inline char* match(char *ptr)
/* Blake Watters 6/18/2002 -- sbw@email.unc.edu */
/* Just a small change, really ;) */
{
	char *str_ptr;
                
   if (strstr(ptr, "vdpop3d") != NULL ) {	
   	str_ptr = strstr(ptr, "logged in from");
                       		 			
      if (str_ptr != NULL) {
      	str_ptr = strstr(str_ptr, "(");
         str_ptr++;
         
         return str_ptr;
 		}
	}

   return NULL;
}
#endif

#ifdef COURIER
inline char* match(char *ptr)
/* This function should match for Courier impad or pop3d, both ssl and non-ssl */
/* log entry submitted by Lang 7/18/2002 -- lang@dominion.tharin.com           */
/* Additional code for various permutations of Courier's log output            */
/* provided by Andreas Grundler -- ag@cronon.org       and                     */
/*             Scott Bronson    -- bronson@rinspin.com                         */
{
	char *str_ptr;
                
	/* Determine that this is an imapd or pop3d entry */
   if ((strstr(ptr, "imapd") != NULL) || ((strstr(ptr, "pop3d") != NULL)) ) {	
   		/* Check that it's not a failure */
   		if (strstr(ptr, "FAILED")) 
   			return NULL;
   			
		/* Check that it's an authenticated entry */		
		if (strstr(ptr, "LOGIN")) {
			str_ptr = strstr(ptr, "ip=[");
			
			/* Skip to the first digit */
			while( !isdigit( (int)(*(++str_ptr)) ) );

			return str_ptr;
		}
	}
	
   return NULL;
}
#endif

#ifdef TPOP3D
inline char* match(char *ptr)
/* Submitted by Alessandro Gatti 7/18/2002 */
/* gatti@kdnr.com */
{
	char *str_ptr;

   if (strstr(ptr, "tpop3d") != NULL) {
		str_ptr = strstr(ptr, "): successfully authenticated");

      if (str_ptr != NULL) {
			str_ptr = strstr(ptr, "(");

         return str_ptr++;
      }
   }

   return NULL;
}
#endif                 

#ifdef CYRUS
inline char* match(char *ptr)
/* Submitted by Laurens van Alphen 8/16/2002 */
/* laurens.van.alphen@keenondots.com         */
{
	char *str_ptr;
	
	/* Determine that this is an imapd or pop3d entry */
	if ((strstr(ptr, "imapd") != NULL) || ((strstr(ptr, "pop3d") != NULL)) ) {
	
		/* Check that it's an authenticated entry */
		if (strstr(ptr, " login: ")) {
			/* Locate the second [ as the first is the pid */
			str_ptr = strstr(ptr, "[");
			str_ptr++;
			
			str_ptr = strstr(str_ptr, "[");
			str_ptr++;
			
			return str_ptr;
		}
	}
	
	return NULL;
}
#endif


void remove_newline(char *ptr)
/* removes the '\n' and any trailing characters */  
{
	while (isdigit(*ptr) || *ptr == '.') 
		ptr++;
		
	*ptr = '\x0';
}

void create_db()
{
	FILE *fp = NULL;
	char filename[160];
	int x = tail;
	
	strcpy(filename, pop_before_smtp_dir);
	strcat(filename, "/pop-before-smtp");
	if (debug == 2)
		strcat(filename, ".txt");

	if (debug == 1) {
		fp = stdout;
	}
	else	
		fp = fopen(filename, "w");
	
	if (fp == NULL) {
		printf("Error, couldn't open output file\n");
		exit(1);
	}
	
	do {
		if (debug)
			printf("x= %d  ", x); 
		if (debug == 2)
			printf("%s OK\n", pop_ok[x].ip_addr);
			
		fprintf(fp, "%s OK\n", pop_ok[x].ip_addr);
		x++;
		x = x % NUM_ENTRIES;
	}
	while (x != head);
	
	if (!debug || debug==2)
		fclose(fp);
}

void purge_list(time_t cur_time)
{
#ifdef OPTIMIZE
	int x, count = 0, local_rewrite;
#endif

	for(;;) {
		if (pop_ok[tail].expire < cur_time) {
			if (debug)
				printf("purging: %ld  %s\n", pop_ok[tail].expire, pop_ok[tail].ip_addr);
				
#ifdef OPTIMIZE
			local_rewrite = 1;
			
			for(x=head; x != tail;x--)
			{
				count++;
				
				if (x < 0)
					x = NUM_ENTRIES;
					
				if (strcmp(pop_ok[tail].ip_addr, pop_ok[x].ip_addr) == 0)
				{
					if (debug) 
						printf("Found OPTIMIZE match, no file output needed\n");
			
					local_rewrite = 0;	
					break;
				}
				
				if (count == SCAN_DEPTH)
					break;
			}
			
			rewrite += local_rewrite;
#endif			
			
			tail++;
			tail = tail % NUM_ENTRIES;
		}
		else 
			break;
		if (tail == head)
			break;
	}
	
	/* Set the time for the alarm to the next expiration */
	alarm(pop_ok[tail].expire - cur_time + 1);		
			
	if (debug) {
		printf("cur_time: %ld   oldest_entry_time: %ld\n", cur_time, pop_ok[tail].expire);
		printf("Alarm set for %ld seconds\n\n", pop_ok[tail].expire - (long int)cur_time);
	}
}

void add_to_list(time_t cur_time, char *addr)
{
#ifdef OPTIMIZE
	int local_rewrite = 1, count = 0, x;
#endif

	remove_newline(addr);
	strcpy(pop_ok[head].ip_addr, addr);
	pop_ok[head].expire = cur_time + GRACE;
	
	head++;
	head = head % NUM_ENTRIES;
	
	if (head == tail) {
		tail++;
		tail = tail % NUM_ENTRIES;
	}
	
#ifdef OPTIMIZE
	x = head - 1;

	for(;x != tail;) {
		x--;
		
		if (x < 0)
			x += NUM_ENTRIES;
		count++;			
		
		if (strcmp(addr, pop_ok[x].ip_addr) == 0)
		{
			if (debug) {
				printf("New entry already in database, no file output needed\n");
				printf("  head-1=%s   x=%s\n", pop_ok[head-1].ip_addr, pop_ok[x].ip_addr);
			}
			
			local_rewrite = 0;	
			break;
		}
				
		if (count == SCAN_DEPTH)
			break;
	}
			
	rewrite += local_rewrite;
#endif				
}

void process_log()
{
	char buffer[1024], *str_ptr;
	time_t lt;
	
	/* current time in seconds */
	lt = time(NULL);
	
	if (debug) {
		/* This printout happens only once */
		strcpy(buffer, "postmap -r ");
		strcat(buffer, pop_before_smtp_dir);
		strcat(buffer, "/pop-before-smtp");
		if (debug == 2)
			strcat(buffer, ".txt");
	
		printf("Postmap command: %s\n", buffer);	
	}
	
	while (fgets(buffer, sizeof(buffer), stdin) != NULL) {
		/* Main loop of the program */

		str_ptr = match(buffer);
		
		if (str_ptr != NULL) {	/* found a vm-pop3d line */
			/* current time in seconds */
			lt = time(NULL);				
			
#ifdef OPTIMIZE			
			/* Start off with the assumption that there will be no re-writes */
			rewrite = 0;			
#endif
			
			add_to_list(lt, str_ptr);
			
			purge_list(lt);
			
#ifdef OPTIMIZE						
			if (rewrite) {
#endif			
			create_db();
			
#ifdef OPTIMIZE
			}
#endif

			if (debug)			
				printf("head=%d tail=%d %ld from %s\n\n", head, tail, lt, str_ptr);
				
			if (!debug || debug==2) {
				strcpy(buffer, "postmap -r ");
				strcat(buffer, pop_before_smtp_dir);
				strcat(buffer, "/pop-before-smtp");									
				if (debug == 2)
					strcat(buffer, ".txt");
			
#ifdef OPTIMIZE
			if (rewrite) {
#endif			
				system(buffer);
#ifdef OPTIMIZE
			}
			else {
				if (debug)
					printf("No rewrite necessary\n");
			}
#endif
			}
		}
	}
}

void fork_off()
{
	int fd[2];
	char buffer[1024];

	/* Here we perform the fork to start tail */
	/* and open a pipe to pop-before-smtp     */
	if (pipe(fd) < 0)
		printf("pipe error");

	if ( (child_id = fork()) < 0)
		printf("fork error");
		
	else if (child_id > 0) {	/* Parent */

		if (fd[0] != STDIN_FILENO) {
			if (dup2(fd[0], STDIN_FILENO) != STDIN_FILENO)
				printf("dup2 error to stdin");
			close(fd[0]);	/* don't need this after dup2 */
		}		
	}
	else {	/* Child will become tail -f */
		strcpy(buffer, "tail -f ");
		strcat(buffer, POP_LOG);
		
		if (debug) {
			printf("Starting command in 3 seconds: %s\n", buffer);
			sleep(3);
		}	
	
		if (fd[0] != STDIN_FILENO) {
			if (dup2(fd[0], STDIN_FILENO) != STDIN_FILENO)
				printf("dup2 error to stdin");
			close(fd[0]);	/* don't need this after dup2 */
		}
		if (fd[1] != STDOUT_FILENO) {
			if (dup2(fd[1], STDOUT_FILENO) != STDOUT_FILENO)
				printf("dup2 error to stdout");
			close(fd[1]);	/* don't need this after dup2 */
		}		

		execlp("tail", "tail", "-f", POP_LOG, (char *) 0);
	}
}

void sig_alarm(int signo)
{
	time_t lt;

	/* current time in seconds */
	lt = time(NULL);				
	
	if (debug) 
		printf("Caught alarm at %ld, going to purge list\n", lt);	
			
	purge_list(lt);
	
	signal(SIGALRM, sig_alarm);
	
	/* Set the time for the alarm to the next expiration */
	alarm(pop_ok[tail].expire - lt + 1);
}

void sig_handler(int signo)
{
	int status;
	
	if (debug) {
		printf("caught signal\n");
	
		printf("going to kill the child (tail) process %d\n", child_id);
	}
	
	if (kill(child_id, SIGTERM) == 0) {
		if (debug)
			printf("kill went ok\n");
		
		if (wait(&status) != child_id) 
			printf("wait error");
	}
	if (debug)  {
		printf("child dead\n");
		sleep(1);
	}
	
	fork_off();
}

int main(int argc, char *argv[])
{
	int x;

	for (x=0;x<argc;x++) {
		if (strcmp(argv[x], "-h") == 0 || strcmp(argv[x], "-?") == 0 ||
			strcmp(argv[x], "-H") == 0) {
			printf("pop-before-smtp in C   version %s\n", VERSION);
			printf("Options:\n");
			printf("  -d   debug, don't actually write\n");
			printf("  -dd  debug, write data to current directory and run postmap command\n");
			exit(0);
		}
		if (strcmp(argv[x], "-v") == 0 || strcmp(argv[x], "-?") == 0 ||
			strcmp(argv[x], "-V") == 0) {
			printf("pop-before-smtp in C   version %s\n", VERSION);
			exit(0);
		}
		if (strcmp(argv[x], "-d") == 0) {
			printf("pop-before-smtp in C   version %s\n", VERSION);		
			printf("Performing debug mode\n");
			debug = 1;
		}
		if (strcmp(argv[x], "-dd") == 0) {
			printf("pop-before-smtp in C   version %s\n", VERSION);		
			printf("Performing debug mode, level 2\n");
			strcpy(pop_before_smtp_dir, ".");
			debug = 2;
		}		
	}

	/* Setup signal handlers */	
	signal(SIGHUP, sig_handler);
	signal(SIGALRM, sig_alarm);	
	
	fork_off();

	process_log();

	return 0;
}
	
